@@ -12,6 +12,8 @@ import java.util.Collections;
 import java.util.Comparator;
 import java.util.Iterator;
 import java.util.List;
+import java.util.function.Function;
+import java.util.function.Supplier;
 import java.util.logging.Level;
 import java.util.logging.Logger;
 import junit.framework.AssertionFailedError;
@@ -21,6 +23,7 @@ import uk.ac.starlink.ecsv.EcsvTableBuilder;
 import uk.ac.starlink.ecsv.EcsvTableWriter;
 import uk.ac.starlink.feather.FeatherTableBuilder;
 import uk.ac.starlink.feather.FeatherTableWriter;
+import uk.ac.starlink.fits.AbstractFitsTableWriter;
 import uk.ac.starlink.fits.AbstractWideFits;
 import uk.ac.starlink.fits.BintableStarTable;
 import uk.ac.starlink.fits.ColFitsTableWriter;
@@ -31,6 +34,8 @@ import uk.ac.starlink.fits.FitsTableWriter;
 import uk.ac.starlink.fits.HealpixFitsTableWriter;
 import uk.ac.starlink.fits.VariableFitsTableWriter;
 import uk.ac.starlink.fits.WideFits;
+import uk.ac.starlink.parquet.ParquetTableBuilder;
+import uk.ac.starlink.parquet.ParquetTableWriter;
 import uk.ac.starlink.table.storage.AdaptiveByteStore;
 import uk.ac.starlink.table.storage.ByteStoreRowStore;
 import uk.ac.starlink.table.storage.FileByteStore;
@@ -95,6 +100,7 @@ public class FormatsTest extends TableCase {
         Logger.getLogger( "uk.ac.starlink.votable" ).setLevel( Level.WARNING );
         Logger.getLogger( "uk.ac.starlink.feather" ).setLevel( Level.SEVERE );
         Logger.getLogger( "uk.ac.starlink.ecsv" ).setLevel( Level.SEVERE );
+        Logger.getLogger( "uk.ac.starlink.parquet" ).setLevel( Level.WARNING );
 
         FitsConstants.configureHierarch();
     }
@@ -177,6 +183,10 @@ public class FormatsTest extends TableCase {
         }
     }
 
+    public void testTransferable() {
+        assertNotNull( new StarTableOutput().getTransferWriter() );
+    }
+
     public void testIdentity() throws IOException {
         checkStarTable( table );
         assertTableEquals( table, table );
@@ -202,7 +212,7 @@ public class FormatsTest extends TableCase {
         StarTableFactory tfact = new StarTableFactory();
         String[] bnames = new String[] {
             "fits", "colfits-basic", "colfits-plus",
-            "votable", "cdf", "ecsv", "feather",
+            "votable", "cdf", "ecsv", "parquet", "feather",
         };
         for ( String bname : bnames ) {
             assertNotNull( tfact.getTableBuilder( bname ) );
@@ -266,6 +276,8 @@ public class FormatsTest extends TableCase {
             "VOTable",
             "CDF",
             "ECSV",
+            "MRT",
+            "parquet",
             "feather",
             "GBIN",
         };
@@ -277,6 +289,8 @@ public class FormatsTest extends TableCase {
             "VOTable",
             "CDF",
             "ECSV",
+            "MRT",
+            "parquet",
             "feather",
             "GBIN",
             "ASCII",
@@ -329,6 +343,7 @@ public class FormatsTest extends TableCase {
             "colfits-basic",
             "votable",
             "ecsv",
+            "parquet",
             "feather",
             "text",
             "ascii",
@@ -410,9 +425,9 @@ public class FormatsTest extends TableCase {
     }
 
     private void exerciseFits( WideFits wide ) throws IOException {
-        boolean allowSignedByte = true;
-        StarTableWriter writer =
-            new FitsTableWriter( "fits", allowSignedByte, wide );
+        FitsTableWriter writer = new FitsTableWriter();
+        writer.setAllowSignedByte( true );
+        writer.setWide( wide );
         File loc = getTempFile( "t.fits" );
         StarTable t1 = table;
         writer.writeStarTable( t1, loc.toString(), sto );
@@ -468,8 +483,11 @@ public class FormatsTest extends TableCase {
             throws IOException {
         File loc = getTempFile( "tv.fits" );
         StarTable t1 = table;
-        new VariableFitsTableWriter( Boolean.valueOf( isLong ), true, wide )
-           .writeStarTable( t1, loc.toString(), sto );
+        VariableFitsTableWriter vWriter =
+            createVariableFitsTableWriter( isLong );
+        vWriter.setAllowSignedByte( true );
+        vWriter.setWide( wide );
+        vWriter.writeStarTable( t1, loc.toString(), sto );
         StarTable t2 = new FitsTableBuilder( wide )
                       .makeStarTable( new FileDataSource( loc ), true,
                                       StoragePolicy.PREFER_MEMORY );
@@ -495,39 +513,37 @@ public class FormatsTest extends TableCase {
     }
 
     public void testReadWrite() throws IOException {
-        for ( WideFits wide : wides_ ) {
-            exerciseReadWrite( new FitsTableWriter( "fits", true, wide ),
-                               new FitsTableBuilder( wide ), "fits" );
-            exerciseReadWrite( new FitsTableWriter( "fits", false, wide ),
-                               new FitsTableBuilder( wide ), "fits" );
-            exerciseReadWrite( new FitsPlusTableWriter( "fits-plus", wide ),
-                               new FitsPlusTableBuilder( wide ), "fits" );
-            exerciseReadWrite( new ColFitsTableWriter( "colfits", wide ),
-                               new ColFitsTableBuilder( wide ), "fits" );
-            exerciseReadWrite(
-                new ColFitsPlusTableWriter( "colfits-plus", wide ),
-                new ColFitsPlusTableBuilder( wide ), "votable" );
-    
-            exerciseReadWrite(
-                new VariableFitsTableWriter( Boolean.FALSE, true, wide ),
-                new FitsTableBuilder( wide ), "fitsv" );
-            exerciseReadWrite(
-                new VariableFitsTableWriter( Boolean.FALSE, false, wide ),
-                new FitsTableBuilder( wide ), "fitsv" );
-    
-            exerciseReadWrite(
-                new VariableFitsTableWriter( Boolean.TRUE, true, wide ),
-                new FitsTableBuilder( wide ), "fitsv" );
-            exerciseReadWrite(
-                new VariableFitsTableWriter( Boolean.TRUE, false, wide ),
-                new FitsTableBuilder( wide ), "fitsv" );
-        }
+
+        exerciseFitsReadWrite( () -> new FitsTableWriter(),
+                               wide -> new FitsTableBuilder( wide ), "fits" );
+        exerciseFitsReadWrite( () -> new ColFitsTableWriter(),
+                               wide -> new ColFitsTableBuilder( wide ), "fits");
+        exerciseFitsReadWrite( () -> createVariableFitsTableWriter( false ),
+                               wide -> new FitsTableBuilder( wide ), "fitsv" );
+        exerciseFitsReadWrite( () -> createVariableFitsTableWriter( true ),
+                               wide -> new FitsTableBuilder( wide ), "fitsv" );
+
+        exerciseReadWrite( new FitsPlusTableWriter(),
+                           new FitsPlusTableBuilder( null ), "votable" );
+        exerciseReadWrite( new ColFitsPlusTableWriter(),
+                           new ColFitsPlusTableBuilder( null ), "votable" );
+
         exerciseReadWrite( new VOTableWriter(),
                            new VOTableBuilder(), "votable" );
         exerciseReadWrite( EcsvTableWriter.SPACE_WRITER,
                            new EcsvTableBuilder(), "ecsv" );
         exerciseReadWrite( EcsvTableWriter.COMMA_WRITER,
                            new EcsvTableBuilder(), "ecsv" );
+
+        ParquetTableBuilder parquetBuilder = new ParquetTableBuilder();
+        ParquetTableWriter parquetWriter = new ParquetTableWriter();
+        parquetBuilder.setCacheCols( Boolean.FALSE );
+        parquetWriter.setGroupArray( false );
+        exerciseReadWrite( parquetWriter, parquetBuilder, "parquet");
+        parquetBuilder.setCacheCols( Boolean.TRUE );
+        parquetWriter.setGroupArray( true );
+        exerciseReadWrite( parquetWriter, parquetBuilder, "parquet");
+
         exerciseReadWrite(
             new FeatherTableWriter( false, StoragePolicy.PREFER_MEMORY ),
             new FeatherTableBuilder(), "feather" );
@@ -544,6 +560,23 @@ public class FormatsTest extends TableCase {
                            new TstTableBuilder(), "text" );
     }
 
+    private void exerciseFitsReadWrite(
+                Supplier<AbstractFitsTableWriter> writerSupplier,
+                Function<WideFits,TableBuilder> builderFunction,
+                String equalMethod )
+            throws IOException {
+        for ( WideFits wide : wides_ ) {
+            TableBuilder builder = builderFunction.apply( wide );
+            for ( int ib = 0; ib < 2; ib++ ) {
+                boolean allowSignedByte = ib == 0;
+                AbstractFitsTableWriter writer = writerSupplier.get();
+                writer.setWide( wide );
+                writer.setAllowSignedByte( allowSignedByte );
+                exerciseReadWrite( writer, builder, equalMethod );
+            }
+        }
+    }
+
     public void exerciseReadWrite( StarTableWriter writer,
                                    TableBuilder reader, String equalMethod )
             throws IOException {
@@ -563,7 +596,7 @@ public class FormatsTest extends TableCase {
                     throw new AssertionError();
             }
             DataSource datsrc =
-                isSeq ? (DataSource) new URLDataSource( loc.toURL() )
+                isSeq ? (DataSource) new URLDataSource( loc.toURI().toURL() )
                       : (DataSource) new FileDataSource( loc );
             StarTable t2 =
                 reader.makeStarTable( datsrc, true,
@@ -594,6 +627,9 @@ public class FormatsTest extends TableCase {
         else if ( "ecsv".equals( equalMethod ) ) {
             assertEcsvTableEquals( t1, t2 );
         }
+        else if ( "parquet".equals( equalMethod ) ) {
+            assertParquetTableEquals( t1, t2 );
+        }
         else if ( "feather".equals( equalMethod ) ) {
             assertFeatherTableEquals( t1, t2 );
         }
@@ -653,7 +689,8 @@ public class FormatsTest extends TableCase {
     }
 
     private void exerciseColFits( WideFits wide ) throws IOException {
-        StarTableWriter writer = new ColFitsTableWriter( "colfits", wide );
+        ColFitsTableWriter writer = new ColFitsTableWriter();
+        writer.setWide( wide );
         File loc = getTempFile( "t.colfits" );
         StarTable t1 = table;
         writer.writeStarTable( t1, loc.toString(), sto );
@@ -848,17 +885,17 @@ public class FormatsTest extends TableCase {
 
     private void assertEcsvTableEquals( StarTable t1, StarTable t2 )
             throws IOException {
-        IntList icols = new IntList();
-        int nc = t1.getColumnCount();
-        for ( int ic = 0; ic < nc; ic++ ) {
-            Class<?> clazz = t1.getColumnInfo( ic ).getContentClass();
-            if ( clazz.getComponentType() == null ) {
-                icols.add( ic );
+        StarTable t1a = new MetaCopyStarTable( t1 );
+        int ncol = t1a.getColumnCount();
+        for ( int ic = 0; ic < ncol; ic++ ) {
+            ColumnInfo cinfo = t1a.getColumnInfo( ic );
+            int[] shape = cinfo.getShape();
+            if ( shape != null &&
+                 shape.length > 1 &&
+                 shape[ shape.length - 1 ] < 0 ) {
+                cinfo.setShape( new int[] { -1 } );
             }
         }
-        StarTable t1a = new ColumnPermutedStarTable( t1, icols.toIntArray() );
-        int ncol = t1a.getColumnCount();
-        assertEquals( ncol, t2.getColumnCount() );
         for ( int ic = 0; ic < ncol; ic++ ) {
             assertValueInfoEquals( t1a.getColumnInfo( ic ),
                                    t2.getColumnInfo( ic ) );
@@ -866,6 +903,19 @@ public class FormatsTest extends TableCase {
         assertRowSequenceEquals( t1a, t2 );
     }
 
+    private void assertParquetTableEquals( StarTable t1, StarTable t2 )
+            throws IOException {
+        int nc = t1.getColumnCount();
+        assertEquals( nc, t2.getColumnCount() );
+        for ( int ic = 0; ic < nc; ic++ ) {
+            ColumnInfo cinfo1 = t1.getColumnInfo( ic );
+            ColumnInfo cinfo2 = t2.getColumnInfo( ic );
+            assertEquals( cinfo1.getName(), cinfo2.getName() );
+            assertEquals( cinfo1.getContentClass(), cinfo2.getContentClass() );
+        }
+        assertRowSequenceEquals( t1, t2 );
+    }
+
     private void assertFeatherTableEquals( StarTable t1, StarTable t2 )
             throws IOException {
         IntList icols = new IntList();
@@ -951,4 +1001,77 @@ public class FormatsTest extends TableCase {
             lrow++;
         }
     }
+
+    private static VariableFitsTableWriter
+            createVariableFitsTableWriter( boolean longIndexing ) {
+        VariableFitsTableWriter vWriter = new VariableFitsTableWriter();
+        vWriter.setStoragePolicy( StoragePolicy.PREFER_MEMORY );
+        vWriter.setLongIndexing( Boolean.valueOf( longIndexing ) );
+        return vWriter;
+    }
+
+    /**
+     * Adjusts a table so that, as far as possible, any variable-length
+     * multi-dimensional arrays are described in the column metadata
+     * as fixed-length multi-dimensional arrays.
+     * This is a utility function to assist with testing formats like ECSV
+     * that don't cope with variable-length multi-dimensional array-valued
+     * columns.
+     *
+     * @param   table   input table
+     * @return   output table
+     */
+    private static StarTable fixArraySizes( StarTable table )
+            throws IOException {
+        table = new MetaCopyStarTable( table );
+        int ncol = table.getColumnCount();
+        int[] nels = new int[ ncol ];
+        RowSequence rseq = table.getRowSequence();
+        ColumnInfo[] cinfos = Tables.getColumnInfos( table );
+        while ( rseq.next() ) {
+            Object[] row = rseq.getRow();
+            for ( int ic = 0; ic < ncol; ic++ ) {
+                if ( cinfos[ ic ].getShape() != null ) {
+                    Object cell = row[ ic ];
+                    if ( cell != null ) {
+                        int nel = Array.getLength( cell );
+                        if ( nels[ ic ] == -1 ) {
+                            // no good
+                        }
+                        else if ( nels[ ic ] == 0 ) {
+                            nels[ ic ] = nel;
+                        }
+                        else if ( nel != nels[ ic ] ) {
+                            nels[ ic ] = -1;
+                        }
+                    }
+                }
+            }
+        }
+        rseq.close();
+        for ( int ic = 0; ic < ncol; ic++ ) {
+            ColumnInfo cinfo = cinfos[ ic ];
+            int[] shape0 = cinfo.getShape();
+            if ( shape0 != null &&
+                 shape0.length > 1 &&
+                 shape0[ shape0.length - 1 ] < 0 ) {
+                int nd0 = shape0.length;
+                int[] shape1;
+                if ( nels[ ic ] >= 1 ) {
+                    shape1 = new int[ nd0 ];
+                    int nel = 1;
+                    for ( int id = 0; id < nd0 - 1; id++ ) {
+                        nel *= shape0[ id ];
+                        shape1[ id ] = shape0[ id ];
+                    }
+                    shape1[ nd0 - 1 ] = nels[ ic ] / nel;
+                }
+                else {
+                    shape1 = new int[] { -1 };
+                }
+                cinfo.setShape( shape1 );
+            }
+        }
+        return table;
+    }
 }
